from abc import ABC
from typing import Any, Dict, List, Optional, Tuple
from latestos.client.base import BaseClient


BROWSER_TIMEOUT = 60


class SeleniumClient(BaseClient, ABC):
    """
    Uses selenium and a selenium driver.
    """
    headers = {}
    timeout = None if BROWSER_TIMEOUT == 0 else BROWSER_TIMEOUT

    def __init__(self) -> None:
        super().__init__()
        self.setup_options()
        self.init_browser()

    def setup_options(self) -> None:
        """
        Setup driver options. Must be overriden.
        """
        raise NotImplementedError()

    def init_browser(self) -> None:
        """
        Initilizes a browser instance.
        """
        raise NotImplementedError()

    def get(self, url: str, *args, **kwargs) -> Tuple[Any, str]:
        """
        Makes a GET request to the given URL.

        Args:
            url (str): url to get

        Returns:
            (Tuple[Any, str]): response object, raw response text
        """
        self.driver.get(url)

        # Get the HTML content of the website
        html_source = self.driver.page_source

        if kwargs.get("json", False):
            # If json data is expected, return raw data and parsed json
            parsed_json = self._parse_json_data(html_source)
            return None, parsed_json
        else:
            return None, html_source

        return None, html_source

    def post(
        self,
        url: str,
        data: dict,
        content_type: Optional[str] = None
    ) -> Tuple[Any, str]:
        """
        Makes a POST request to the given URL.

        Args:
            url (str): url to post to
            data (dict): data to send
            content_type (str): content type of the request

        Returns:
            (Tuple[Any, str]): response object, raw response text
        """
        raise NotImplementedError()

    def get_cookies(self) -> List[Dict[str, str]]:
        """
        Get cookies in use.

        Returns:
            (List[Dict[str, str]]): cookies - each of 'em is a dict {name: val}
        """
        cookies = self.driver.get_cookies()
        formatted_cookies = [{c.get("name"): c.get("value")} for c in cookies]
        return formatted_cookies

    def set_cookies(self, cookies: list, url: str):
        """
        Set cookies.

        Args:
            cookies (List[Dict[str, str]]): cookies - each is a dict {name: val}
            url (str): domain's base url
        """
        for cookie in cookies:
            cookie_data = cookie.items()
            for cookie_name, cookie_val in cookie_data:
                self.driver.add_cookie({
                    "name": cookie_name,
                    "value": cookie_val,
                    "domain": url
                })

    def close(self) -> None:
        """
        Closes the client / ends sessions.
        """
        self.driver.close()
        self.driver.quit()

    def restart(self):
        """
        Restarts the client
        """
        self.close()
        self.init_browser()

    def _get_page_source(self, *args) -> str:
        """
        Get the raw page source.

        Returns:
            (str): html page source
        """
        try:
            return self.driver.page_source
        except Exception:
            return ""



from abc import ABC
from typing import Any, Dict, List, Tuple


class BaseClient(ABC):
    """ Defines the interface used to make requests through the web """

    def get(self, url: str, *args, **kwargs) -> Tuple[Any, str]:
        """
        Makes a GET request to the given URL.

        Args:
            url (str): url to get

        Returns:
            (Tuple[Any, str]): response object, raw response text
        """
        raise NotImplementedError()

    def post(self, url: str, data: dict, content_type: str) -> Tuple[Any, str]:
        """
        Makes a POST request to the given URL.

        Args:
            url (str): url to post to
            data (dict): data to send
            content_type (str): content type of the request

        Returns:
            (Tuple[Any, str]): response object, raw response text
        """
        raise NotImplementedError()

    def get_cookies(self) -> List[Dict[str, str]]:
        """
        Get cookies in use.

        Returns:
            (List[Dict[str, str]]): cookies - each of 'em is a dict {name: val}
        """
        raise NotImplementedError()

    def set_cookies(self, cookies: list, url: str):
        """
        Set cookies.

        Args:
            cookies (List[Dict[str, str]]): cookies - each is a dict {name: val}
            url (str): domain's base url
        """
        raise NotImplementedError()

    def close(self):
        """
        Closes the client / ends sessions.
        """
        raise NotImplementedError()

    def restart(self):
        """
        Restarts the client
        """
        raise NotImplementedError()



import os
import sys
from selenium import webdriver
from selenium.webdriver import FirefoxProfile
from selenium.webdriver.firefox.options import Options
from selenium.common.exceptions import WebDriverException
from latestos.client.selenium.base import SeleniumClient


class FirefoxClient(SeleniumClient):
    """
    Uses firefox through Selenium.
    """
    PROFILE_PREFERENCES = {
        "network.http.pipelining": True,
        "network.http.proxy.pipelining": True,
        "network.http.pipelining.maxrequests": 8,
        "content.notify.interval": 500000,
        "content.notify.ontimer": True,
        "content.switch.threshold": 250000,
        "browser.cache.memory.capacity": 65536, # Increase the cache capacity
        "browser.startup.homepage": "about:blank",
        "reader.parse-on-load.enabled": False, # Disable reader
        "browser.pocket.enabled": False, # Duck pocket
        "loop.enabled": False,
        "browser.chrome.toolbar_style": 1, # Text on Toolbar instead of icons
        "browser.display.show_image_placeholders": False, # No thumbnails
        "browser.display.use_document_colors": False, # No document colors
        "browser.display.use_document_fonts": 0, # Don't load document fonts
        "browser.display.use_system_colors": True, # Use system colors
        "browser.formfill.enable": False, # Autofill on forms disabled
        "browser.helperApps.deleteTempFileOnExit": True, # Delete temp files
        "browser.shell.checkDefaultBrowser": False,
        "browser.startup.homepage": "about:blank",
        "browser.startup.page": 0, # blank
        "browser.tabs.forceHide": True, # Disable tabs
        "browser.urlbar.autoFill": False, # Disable autofill on URL bar
        "browser.urlbar.autocomplete.enabled": False, # No autocomplete on bar
        "browser.urlbar.showPopup": False, # No list of URLs when typing on bar
        "browser.urlbar.showSearch": False, # Disable search bar
        "extensions.checkCompatibility": False, # Addon update disabled
        "extensions.checkUpdateSecurity": False,
        "extensions.update.autoUpdateEnabled": False,
        "extensions.update.enabled": False,
        "general.startup.browser": False,
        "plugin.default_plugin_disabled": False,
        "permissions.default.image": 2, # Image load disabled again
    }

    def setup_options(self) -> None:
        """
        Setup driver options.
        """
        self.options = Options()
        self.options.headless = True
        self.options.log.level = "TRACE"

        self.profile = FirefoxProfile()
        for pref_name, pref_val in self.__class__.PROFILE_PREFERENCES.items():
            self.profile.set_preference(pref_name, pref_val)

    def init_browser(self) -> None:
        """
        Initilizes a browser instance.
        """
        try:
            self.driver = webdriver.Firefox(
                self.profile,
                options=self.options,
                service_log_path=os.path.devnull
            )
        except WebDriverException:
            print("geckodriver executable needs to be in path")
            sys.exit(1)



import re
import sys
from typing import Tuple
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support.select import Select
from latestos.scraper.base import BaseScraper
from latestos.client.selenium.firefox import FirefoxClient
from latestos.client.selenium.base import BROWSER_TIMEOUT as TIMEOUT


ISO_DOWNLOAD = "https://www.microsoft.com/en-us/software-download/windowsinsi" \
               "derpreviewiso?wa=wsignin1.0"


SELECTORS = {
    # Usually appears when not logged in yet
    "PROGRESS": "#progressLoadingMessage",
    # Appears when submitting edition options
    "ANOTHER_PROGRESS": "#progressModalMessage",
    "ISO_DOWNLOAD_HOME_LOGIN_BTN": "#mectrl_main_trigger",
    "LOGIN_PAGE_EMAIL": "#i0116",
    "LOGIN_PAGE_EMAIL_NEXT": "#idSIButton9",
    "LOGIN_PAGE_EMAIL_ERROR": "#usernameError",
    "PERSONAL_ACCOUNT": "#msaTile",
    "LOGIN_PAGE_PASSWORD": "#i0118",
    "LOGIN_PAGE_PASSWORD_NEXT": "#idSIButton9",
    "LOGIN_PAGE_PASSWORD_ERROR": "#passwordError",
    "DOWNLOAD_PAGE_HEADER": "#headerArea",
    "SELECT_EDITION": "#product-edition",
    "SUBMIT_EDITION": "#submit-product-edition",
    "SELECT_LANGUAGE": "#product-languages",
    "SUBMIT_LANGUAGE": "#submit-sku",
    "DOWNLOAD_URLS": "#card-info-content > div > div > div > a",
    "DOWNLOAD_NAMES": "#card-info-content > div > div > div > a > span",
    "DOWNLOAD_ERROR": "#errorModalTitle",
    "64BIT_DOWNLOAD": "#card-info-content > div > div:nth-child(1) > div > a",
    "ACCOUNT_BUTTON": "#meControl",
    "SIGN_OUT_BUTTON": "#mectrl_body_signOut",
}


class WindowsInsidersPreviewScraper(BaseScraper):
    def get_latest_release_data(self) -> Tuple[str, str, str]:
        """
        Gets latest OS release data.

        Returns:
            (str, str, str): iso url, checksum url, os version
        """
        # Extract data from latest release
        iso_url = self.get_iso_url()
        checksum_url = None
        version = self.get_iso_version(iso_url)

        return iso_url, checksum_url, version

    def get_iso_url(self) -> str:
        client = FirefoxClient()

        self._open_download_page(client)
        self._open_login_page(client)
        self._submit_email(client)
        self._select_personal_account(client)
        self._submit_password(client)
        self._select_product_edition(client)
        self._select_product_language(client)
        iso_url = self._get_download_link(client)

        self._attempt_logout(client)

        return iso_url

    def _open_download_page(self, client: FirefoxClient) -> None:
        try:
            client.get(ISO_DOWNLOAD)
            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.invisibility_of_element_located(
                    (By.CSS_SELECTOR, SELECTORS["PROGRESS"])
                )
            )
        except Exception as e:
            self._handle_client_error(
                client, e, f"Error while loading {ISO_DOWNLOAD}")

    def _open_login_page(self, client: FirefoxClient) -> None:
        try:
            login_button = client.driver.find_element_by_css_selector(
                SELECTORS["ISO_DOWNLOAD_HOME_LOGIN_BTN"])
            login_button.click()

            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, SELECTORS["LOGIN_PAGE_EMAIL"])
                )
            )
        except Exception as e:
            self._handle_client_error(
                client, e, f"Error while loading login form")

    def _submit_email(self, client: FirefoxClient) -> None:
        try:
            valid_email = False
            while not valid_email:
                valid_email = self._enter_and_validate_email(client)
        except Exception as e:
            self._handle_client_error(
                client, e, f"Error while submitting email")

    def _enter_and_validate_email(self, client: FirefoxClient) -> bool:
        email_input = client.driver.find_element_by_css_selector(
                SELECTORS["LOGIN_PAGE_EMAIL"])
        email = input("Enter your Microsoft's account email: ")
        email_input.send_keys(email)

        email_next = client.driver.find_element_by_css_selector(
            SELECTORS["LOGIN_PAGE_EMAIL_NEXT"])
        email_next.click()

        _ = WebDriverWait(client.driver, TIMEOUT).until(
            EC.visibility_of_element_located((
                By.CSS_SELECTOR,
                f"{SELECTORS['LOGIN_PAGE_EMAIL_ERROR']}, " \
                f"{SELECTORS['PERSONAL_ACCOUNT']}, " \
                f"{SELECTORS['LOGIN_PAGE_PASSWORD']}"
            ))
        )

        # Check if there was an error
        try:
            _ = client.driver.find_element_by_css_selector(
                SELECTORS["LOGIN_PAGE_EMAIL_ERROR"])
            return False
        except Exception:
            return True

    def _select_personal_account(self, client: FirefoxClient) -> None:
        try:
            personal_account = client.driver.find_element_by_css_selector(
                SELECTORS["PERSONAL_ACCOUNT"])
            personal_account.click()

            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, SELECTORS["LOGIN_PAGE_PASSWORD"])
                )
            )
        except NoSuchElementException as e:
            return
        except Exception as e:
            self._handle_client_error(
                client, e, f"Error while selecting account")

    def _submit_password(self, client: FirefoxClient) -> None:
        try:
            valid_password = False
            while not valid_password:
                valid_password = self._enter_and_validate_password(client)
        except Exception as e:
            self._handle_client_error(
                client, e, f"Error while submitting password")

    def _enter_and_validate_password(self, client: FirefoxClient) -> bool:
        password_input = client.driver.find_element_by_css_selector(
                SELECTORS["LOGIN_PAGE_PASSWORD"])
        password = input("Enter your Microsoft's account password: ")
        password_input.send_keys(password)

        password_next = client.driver.find_element_by_css_selector(
            SELECTORS["LOGIN_PAGE_PASSWORD_NEXT"])
        password_next.click()

        _ = WebDriverWait(client.driver, TIMEOUT).until(
            EC.visibility_of_element_located((
                By.CSS_SELECTOR,
                f"{SELECTORS['LOGIN_PAGE_PASSWORD_ERROR']}, " \
                f"{SELECTORS['DOWNLOAD_PAGE_HEADER']}"
            ))
        )

        # Check if there was an error
        try:
            _ = client.driver.find_element_by_css_selector(
                SELECTORS["LOGIN_PAGE_PASSWORD_ERROR"])
            return False
        except Exception:
            return True

    def _select_product_edition(self, client: FirefoxClient) -> None:
        try:
            edition = Select(client.driver.find_element_by_css_selector(
                SELECTORS["SELECT_EDITION"]))
            edition.select_by_index(4) # Enterprise Dev Channel
            edition_submit = client.driver.find_element_by_css_selector(
                SELECTORS["SUBMIT_EDITION"])
            edition_submit.click()

            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.element_to_be_clickable(
                    (By.CSS_SELECTOR, SELECTORS["SELECT_LANGUAGE"])
                )
            )
        except Exception as e:
            self._handle_client_error(
                client, e, "Error while selecting product edition")

    def _select_product_language(self, client: FirefoxClient) -> None:
        try:
            language = Select(client.driver.find_element_by_css_selector(
                SELECTORS["SELECT_LANGUAGE"]))
            language.select_by_index(8) # English
            language_submit = client.driver.find_element_by_css_selector(
                SELECTORS["SUBMIT_LANGUAGE"])
            language_submit.click()

            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.visibility_of_element_located((
                    By.CSS_SELECTOR,
                    f"{SELECTORS['64BIT_DOWNLOAD']}, " \
                    f"{SELECTORS['DOWNLOAD_ERROR']}"
                ))
            )

            try:
                _ = client.driver.find_element_by_css_selector(
                    SELECTORS["64BIT_DOWNLOAD"])
            except Exception as e:
                raise e
        except Exception as e:
            self._handle_client_error(
                client, e, "Error while selecting product language")

    def _get_download_link(self, client: FirefoxClient) -> str:
        try:
            link = client.driver.find_element_by_css_selector(
                SELECTORS["64BIT_DOWNLOAD"])
            return link.get_attribute("href")
        except Exception as e:
            self._handle_client_error(
                client, e, "Error while extracting download link")

    def _attempt_logout(self, client: FirefoxClient) -> None:
        try:
            # client.driver.save_screenshot("./download-page.png")
            self._logout(client)
            # client.driver.save_screenshot("./logged-out.png")
            client.close()
        except Exception:
            pass

    def _logout(self, client: FirefoxClient) -> None:
        try:
            # Look for sign out button
            account = client.driver.find_element_by_css_selector(
                SELECTORS["ACCOUNT_BUTTON"])
            account.click()

            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, SELECTORS["SIGN_OUT_BUTTON"])
                )
            )

            # Click sign out button
            sign_out = client.driver.find_element_by_css_selector(
                SELECTORS["SIGN_OUT_BUTTON"])
            sign_out.click()

            # Wait until logged out
            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, SELECTORS["PROGRESS"])
                )
            )
            _ = WebDriverWait(client.driver, TIMEOUT).until(
                EC.invisibility_of_element_located(
                    (By.CSS_SELECTOR, SELECTORS["PROGRESS"])
                )
            )
        except Exception as e:
            self._handle_client_error(
                client, e, f"Error while logging out")

    def _handle_client_error(
        self, client: FirefoxClient, e: Exception, msg: str
    ) -> None:
        """
        Handles a web client error

        Args:
            client ([type]): webclient
            msg (str): additional message
            e (Exception): the raised exception

        Raises:
            Exception: exception
        """
        client.driver.save_screenshot("./error.png")
        client.close()
        print(f"{msg} - {e}")
        sys.exit(1)

    def get_iso_version(self, iso_url: str) -> str:
        """
        Extracts Windows Insiders Preview ISO version from the ISO url

        Args:
            iso_url (str): iso url

        Raises:
            ValueError: error at extracting iso version

        Returns:
            str: iso version
        """
        # This pattern will find the version, preceded by language
        # Example: en-us_19043_928
        pattern = r"(?<=Windows10_InsiderPreview_Client_x\d\d_)(.*)(?=.iso)"

        res = re.findall(pattern, iso_url)

        if not res:
            raise ValueError(f"Could not extract WIP version from {iso_url}")

        version_with_lang = res[0]

        # Extract the number, then replace _ with .
        try:
            return version_with_lang.split("_", 1)[1].replace("_", ".")
        except Exception:
            raise ValueError(
                f"Could not extract WIP version from {version_with_lang}")



from latestos.scraper.base import BaseScraper
from latestos.scraper.arch import ArchScraper
from latestos.scraper.centos import CentOSScraper
from latestos.scraper.fedora import FedoraScraper
from latestos.scraper.ubuntu import UbuntuScraper
from latestos.scraper.debian import DebianScraper
from latestos.scraper.raspbian import RaspbianScraper
from latestos.scraper.windows import WindowsInsidersPreviewScraper


def get_os_scraper(os_name: str) -> BaseScraper:
    """
    Takes an OS name and returns the corresponding scraper.

    Returns:
        (BaseScraper): scraper
    """
    os_name = os_name.lower()

    if os_name == "arch":
        return ArchScraper()
    elif os_name == "centos":
        return CentOSScraper()
    elif os_name == "fedora":
        return FedoraScraper()
    elif os_name == "ubuntu":
        return UbuntuScraper()
    elif os_name == "debian":
        return DebianScraper()
    elif os_name == "raspbian":
        return RaspbianScraper()
    elif os_name == "windows":
        return WindowsInsidersPreviewScraper()
    else:
        raise ValueError(f"{os_name}")
